/*
 * Read_Config.c - Read the overall SLURM configuration file
 * This module also has the basic functions for manipulation of data structures
 * See slurm.h for documentation on external functions and data structures
 *
 * Author: Moe Jette, jette@llnl.gov
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "slurm.h"

#define BUF_SIZE 1024
#define DEBUG_MODULE 1
#define DEBUG_SYSTEM 1
#define NO_VAL (-99)
#define NODE_BUF_SIZE 1024
#define SEPCHARS " \n\t"

void 	BitMapAND(unsigned *BitMap1, unsigned *BitMap2);
void 	BitMapClear(unsigned *BitMap, int Position);
void 	BitMapOR(unsigned *BitMap1, unsigned *BitMap2);
char 	*BitMapPrint(unsigned *BitMap);
void 	BitMapSet(unsigned *BitMap, int Position);
int 	BitMapValue(unsigned *BitMap, int Position);
void	Dump_Hash();
int 	Hash_Index(char *name);
int 	Load_Integer(int *destination, char *keyword, char *In_Line);
int 	Load_String(char **destination, char *keyword, char *In_Line);
int 	Parse_Node_Spec(char *In_Line);
int 	Parse_Part_Spec(char *In_Line);
void 	Rehash();
void	Report_Leftover(char *In_Line, int Line_Num);

struct {
    struct Config_Record Config;
    struct Node_Record Node;
} Default_Record;
struct Part_Record Default_Part;

#if DEBUG_MODULE
/* main is used here for module testing purposes only */
main(int argc, char * argv[]) {
    int Error_Code, Start_Inx, End_Inx, Count_Inx;
    char Out_Line[BUF_SIZE];
    char *Format, *NodeName, *BitMap;
    char *Partitions[] = {"login", "debug", "batch", "class", "END"};
    struct Part_Record *Part_Ptr;
    int i, found;

    if (argc > 2) {
	printf("Usage: %s <in_file>\n", argv[0]);
	exit(0);
    } /* if */

    Error_Code = Init_SLURM_Conf();
    if (Error_Code != 0) exit(Error_Code);

    if (argc == 2) 
	Error_Code = Read_SLURM_Conf(argv[1]);
    else
	Error_Code = Read_SLURM_Conf(SLURM_CONF);

    if (Error_Code != 0) {
	printf("Error %d from Read_SLURM_Conf\n", Error_Code);
	exit(1);
    } /* if */

    printf("ControlMachine=%s\n", ControlMachine);
    printf("BackupController=%s\n", BackupController);
    printf("\n");

    for (i=0; i<Node_Record_Count; i++) {
	if (strlen((Node_Record_Table_Ptr+i)->Name) == 0) continue;
	printf("NodeName=%s ",      (Node_Record_Table_Ptr+i)->Name);
	printf("NodeState=%s ",     Node_State_String[(Node_Record_Table_Ptr+i)->NodeState]);
	printf("LastResponse=%ld ", (long)(Node_Record_Table_Ptr+i)->LastResponse);

	printf("CPUs=%d ",          (Node_Record_Table_Ptr+i)->Config_Ptr->CPUs);
	printf("RealMemory=%d ",    (Node_Record_Table_Ptr+i)->Config_Ptr->RealMemory);
	printf("TmpDisk=%d ",       (Node_Record_Table_Ptr+i)->Config_Ptr->TmpDisk);
	printf("Weight=%d ",        (Node_Record_Table_Ptr+i)->Config_Ptr->Weight);
	printf("Feature=%s\n",        (Node_Record_Table_Ptr+i)->Config_Ptr->Feature);
    } /* for */
    BitMap = BitMapPrint(Up_NodeBitMap);
    printf("\nUp_NodeBitMap  =%s\n", BitMap);
    free(BitMap);
    BitMap = BitMapPrint(Idle_NodeBitMap);
    printf("Idle_NodeBitMap=%s\n\n", BitMap);
    free(BitMap);

    printf("Default_Part_Name=%s\n", Default_Part_Name);
    found = 0;
    for (i=0; ;i++) {
	if (strcmp(Partitions[i], "END") == 0) {
	    if (found) break;
	    Part_Ptr = Default_Part_Loc;
	} else {
	    Part_Ptr = Find_Part_Record(Partitions[i]);
	    if (Part_Ptr == Default_Part_Loc) found = 1;
	} /* else */
	if (Part_Ptr == NULL) continue;
	printf("PartitionName=%s ",     Part_Ptr->Name);
	printf("MaxTime=%d ",           Part_Ptr->MaxTime);
	printf("MaxNodes=%d ",          Part_Ptr->MaxNodes);
	printf("Key=%d ",               Part_Ptr->Key);
	printf("StateUp=%d ",           Part_Ptr->StateUp);
	printf("Nodes=%s ",             Part_Ptr->Nodes);
	printf("TotalNodes=%d ",        Part_Ptr->TotalNodes);
	printf("TotalCPUs=%d ",         Part_Ptr->TotalCPUs);
	printf("AllowGroups=%s  ",      Part_Ptr->AllowGroups);
	BitMap = BitMapPrint(Part_Ptr->NodeBitMap);
	printf("NodeBitMap=%s\n",	BitMap);
	if (BitMap) free(BitMap);
    } /* for */

    exit(0);
} /* main */
#endif



/*
 * BitMapAND - AND two bitmaps together
 * Input: BitMap1 and BitMap2 - The bitmaps to AND
 * Output: BitMap1 is set to the value of BitMap1 & BitMap2
 */
void BitMapAND(unsigned *BitMap1, unsigned *BitMap2) {
    int i, size;

    size = (Node_Record_Count + (sizeof(unsigned)*8) - 1) / (sizeof(unsigned)*8);
    for (i=0; i<size; i++) {
	BitMap1[i] &= BitMap2[i];
    } /* for (i */
} /* BitMapAND */


/*
 * BitMapClear - Clear the specified bit in the specified bitmap
 * Input: BitMap - The bit map to manipulate
 *        Position - Postition to clear
 */
void BitMapClear(unsigned *BitMap, int Position) {
    int val, bit, mask;

    val  = Position / (sizeof(unsigned)*8);
    bit  = Position % (sizeof(unsigned)*8);
    mask = ~(0x1 << ((sizeof(unsigned)*8)-1-bit));

    BitMap[val] &= mask;
} /* BitMapClear */


/*
 * BitMapCopy - AND two bitmaps together
 * Input: BitMap - The bitmap create a copy of
 * Output: Returns pointer to copy of BitMap or NULL if error (no memory)
 *   The returned value MUST BE FREED by the calling routine
 */
unsigned * BitMapCopy(unsigned *BitMap) {
    int i, size;
    unsigned *Output;

    size = (Node_Record_Count + (sizeof(unsigned)*8) - 1) / (sizeof(unsigned)*8);
    size *= sizeof(unsigned);
    Output = malloc(size);
    if (Output == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "BitMapCopy: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "BitMapCopy: unable to allocate memory\n")
#endif
	return NULL;
    } /* if */

    for (i=0; i<size; i++) {
	Output[i] &= BitMap[i];
    } /* for (i */
    return Output;
} /* BitMapCopy */


/*
 * BitMapOR - OR two bitmaps together
 * Input: BitMap1 and BitMap2 - The bitmaps to OR
 * Output: BitMap1 is set to the value of BitMap1 | BitMap2
 */
void BitMapOR(unsigned *BitMap1, unsigned *BitMap2) {
    int i, size;

    size = (Node_Record_Count + (sizeof(unsigned)*8) - 1) / (sizeof(unsigned)*8);
    for (i=0; i<size; i++) {
	BitMap1[i] |= BitMap2[i];
    } /* for (i */
} /* BitMapOR */


/*
 * BitMapPrint - Convert the specified bitmap into a printable hexadecimal string
 * Input: BitMap - The bit map to print
 * Output: Returns a string
 * NOTE: The returned string must be freed by the calling program
 */
char *BitMapPrint(unsigned *BitMap) {
    int i, j, k, size, nibbles;
    char *Output, temp_str[2];

    size = (Node_Record_Count + (sizeof(unsigned)*8) - 1) / (sizeof(unsigned)*8);
    nibbles = (Node_Record_Count + 3) / 4;
    Output = (char *)malloc((sizeof(unsigned)*size*2)+3);
    if (Output == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "BitMapPrint: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "BitMapPrint: unable to allocate memory\n")
#endif
	return NULL;
    } /* if */

    strcpy(Output, "0x");
    k = 0;
    for (i=0; i<size; i++) {
	for (j=((sizeof(unsigned)*8)-4); j>=0; j-=4) {
	    sprintf(temp_str, "%x", ((BitMap[i]>>j)&0xf));
	    strcat(Output, temp_str);
	    k++;
	    if (k == nibbles) return Output;
	} /* for (j */
    } /* for (i */
    return Output;
} /* BitMapPrint */


/*
 * BitMapSet - Set the specified bit in the specified bitmap
 * Input: BitMap - The bit map to manipulate
 *        Position - Postition to set
 */
void BitMapSet(unsigned *BitMap, int Position) {
    int val, bit, mask;

    val  = Position / (sizeof(unsigned)*8);
    bit  = Position % (sizeof(unsigned)*8);
    mask = (0x1 << ((sizeof(unsigned)*8)-1-bit));

    BitMap[val] |= mask;
} /* BitMapSet */


/*
 * BitMapValue - Return the specified bit in the specified bitmap
 * Input: BitMap - The bit map to get value from
 *        Position - Postition to get
 * Output: Returns the value 0 or 1
 */
int BitMapValue(unsigned *BitMap, int Position) {
    int val, bit, mask;

    val  = Position / (sizeof(unsigned)*8);
    bit  = Position % (sizeof(unsigned)*8);
    mask = (0x1 << ((sizeof(unsigned)*8)-1-bit));

    mask &= BitMap[val];
    if (mask == 0)
	return 0;
    else
	return 1;
} /* BitMapValue */


/*
 * Build_BitMaps - Build node bitmaps to define which nodes are in which 
 *    1) Partition  2) Configuration record  3) UP state  4) IDLE state
 *    Also sets values of TotalNodes and TotalCPUs for every partition.
 * Output: Returns 0 if no error, errno otherwise
 */
int Build_BitMaps() {
    int i, j, size;
    ListIterator Config_Record_Iterator;	/* For iterating through Config_Record */
    ListIterator Part_Record_Iterator;		/* For iterating through Part_Record_List */
    struct Config_Record *Config_Record_Point;	/* Pointer to Config_Record */
    struct Part_Record *Part_Record_Point;	/* Pointer to Part_Record */
    struct Node_Record *Node_Record_Point;	/* Pointer to Node_Record */
    unsigned *AllPart_NodeBitMap;
    char *Format, This_Node_Name[BUF_SIZE];
    int Start_Inx, End_Inx, Count_Inx;
    char *str_ptr1, *str_ptr2;

    size = (Node_Record_Count + (sizeof(unsigned)*8) - 1) / 
		(sizeof(unsigned)*8); 	/* Unsigned int records in bitmap */
    size *= 8;				/* Bytes in bitmap */

    /* Scan all nodes and identify which are UP and IDLE */
    if (Idle_NodeBitMap) free(Idle_NodeBitMap);
    if (Up_NodeBitMap)   free(Up_NodeBitMap);
    Idle_NodeBitMap = (unsigned *)malloc(size);
    Up_NodeBitMap   = (unsigned *)malloc(size);
    if ((Idle_NodeBitMap == NULL) || (Up_NodeBitMap == NULL)) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Build_BitMaps: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Build_BitMaps: unable to allocate memory\n")
#endif
	return ENOMEM;
    } /* if */
    memset(Idle_NodeBitMap, 0, size);
    memset(Up_NodeBitMap, 0, size);
    for (i=0; i<Node_Record_Count; i++) {
	if (strlen((Node_Record_Table_Ptr+i)->Name) == 0) continue;	/* Defunct */
	if ((Node_Record_Table_Ptr+i)->NodeState == STATE_IDLE) BitMapSet(Idle_NodeBitMap, i);
	if ((Node_Record_Table_Ptr+i)->NodeState != STATE_DOWN) BitMapSet(Up_NodeBitMap, i);
    } /* for */

    /* Scan configuration records and identify nodes in each */
    Config_Record_Iterator = list_iterator_create(Config_List);
    if (Config_Record_Iterator == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Build_BitMaps: list_iterator_create unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Build_BitMaps: list_iterator_create unable to allocate memory\n");
#endif
	return ENOMEM;
    }
    while (Config_Record_Point = (struct Config_Record *)list_next(Config_Record_Iterator)) {
	if (Config_Record_Point->NodeBitMap) free(Config_Record_Point->NodeBitMap);
	Config_Record_Point->NodeBitMap = (unsigned *)malloc(size);
	if (Config_Record_Point->NodeBitMap == NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Build_BitMaps: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Build_BitMaps: unable to allocate memory\n")
#endif
	    return ENOMEM;
	} /* if */
	memset(Config_Record_Point->NodeBitMap, 0, size);

	/* Check for each node in the configuration record */
	Parse_Node_Name(Config_Record_Point->Nodes, &Format, &Start_Inx, &End_Inx, &Count_Inx);
	for (i=Start_Inx; i<=End_Inx; i++) {
	    if (Count_Inx == 0) {	/* Deal with comma separated node names here */
		if (i == Start_Inx)
		    str_ptr2 = strtok_r(Format, ",", &str_ptr1);
		else
		    str_ptr2 = strtok_r(NULL, ",", &str_ptr1);
		if (str_ptr2 == NULL) break;
		End_Inx++;
		strcpy(This_Node_Name, str_ptr2);
	    } else
		sprintf(This_Node_Name, Format, i);
	    Node_Record_Point = Find_Node_Record(This_Node_Name);
	    if (Node_Record_Point == NULL) continue;
	    if (Node_Record_Point->Config_Ptr != Config_Record_Point) {
	    } /* if */
	    j = Node_Record_Point - Node_Record_Table_Ptr;
	    BitMapSet(Config_Record_Point->NodeBitMap, j);
	} /* for */
    } /* while */
    list_iterator_destroy(Config_Record_Iterator);


    /* Scan partition table and identify nodes in each */
    AllPart_NodeBitMap = (unsigned *)malloc(size);
    if (AllPart_NodeBitMap == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Build_BitMaps: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Build_BitMaps: unable to allocate memory\n")
#endif
	return ENOMEM;
    } /* if */
    memset(AllPart_NodeBitMap, 0, size);
    Part_Record_Iterator = list_iterator_create(Part_List);
    if (Part_Record_Iterator == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Build_BitMaps: list_iterator_create unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Build_BitMaps: list_iterator_create unable to allocate memory\n");
#endif
	return ENOMEM;
    } /* if */
    while (Part_Record_Point = (struct Part_Record *)list_next(Part_Record_Iterator)) {
	if (Part_Record_Point->NodeBitMap) free(Part_Record_Point->NodeBitMap);
	Part_Record_Point->NodeBitMap = (unsigned *)malloc(size);
	if (Part_Record_Point->NodeBitMap == NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Build_BitMaps: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Build_BitMaps: unable to allocate memory\n")
#endif
	    free(AllPart_NodeBitMap);
	    return ENOMEM;
	} /* if */
	memset(Part_Record_Point->NodeBitMap, 0, size);

	/* Check for each node in the partition */
	Parse_Node_Name(Part_Record_Point->Nodes, &Format, &Start_Inx, &End_Inx, &Count_Inx);
	for (i=Start_Inx; i<=End_Inx; i++) {
	    if (Count_Inx == 0) {	/* Deal with comma separated node names here */
		if (i == Start_Inx)
		    str_ptr2 = strtok_r(Format, ",", &str_ptr1);
		else
		    str_ptr2 = strtok_r(NULL, ",", &str_ptr1);
		if (str_ptr2 == NULL) break;
		End_Inx++;
		strcpy(This_Node_Name, str_ptr2);
	    } else
		sprintf(This_Node_Name, Format, i);
	    Node_Record_Point = Find_Node_Record(This_Node_Name);
	    if (Node_Record_Point == NULL) continue;
	    j = Node_Record_Point - Node_Record_Table_Ptr;
	    if (BitMapValue(AllPart_NodeBitMap, j) == 1) {
#if DEBUG_SYSTEM
		fprintf(stderr, "Build_BitMaps: Node %s defined in more than one partition\n", 
			This_Node_Name);
		fprintf(stderr, "Build_BitMaps: Only the first partition's specification is honored\n");
#else
		syslog(LOG_ERR, "Build_BitMaps: Node %s defined in more than one partition\n", 
			This_Node_Name);
		syslog(LOG_ERR, "Build_BitMaps: Only the first partition's specification is honored\n");
#endif
	    } else {
		BitMapSet(Part_Record_Point->NodeBitMap, j);
		Part_Record_Point->TotalNodes++;
		Part_Record_Point->TotalCPUs += Node_Record_Point->Config_Ptr->CPUs;
	    } /* else */
	} /* for */
    } /* while */
    BitMapSet(AllPart_NodeBitMap, j);
   list_iterator_destroy(Part_Record_Iterator);
    return 0;
} /* Build_BitMaps */

/*
 * Create_Config_Record - Create a Config_Record entry, append it to the Config_List, 
 *	and set is values to the defaults.
 * Input: Error_Code - Pointer to an error code
 * Output: Returns pointer to the Config_Record
 *         Error_Code - set to zero if no error, errno otherwise
 * NOTE: The pointer returned is allocated memory that must be freed when no longer needed.
 */
struct Config_Record *Create_Config_Record(int *Error_Code) {
    struct Config_Record *Config_Point;

    Config_Point = (struct Config_Record *)malloc(sizeof(struct Config_Record));
    if (Config_Point == (struct Config_Record *)NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Create_Config_Record: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Create_Config_Record: unable to allocate memory\n")
#endif
	*Error_Code = ENOMEM;
	return (struct Config_Record *)NULL;
    } /* if */

    /* Set default values */
    Config_Point->CPUs = Default_Record.Config.CPUs;
    Config_Point->RealMemory = Default_Record.Config.RealMemory;
    Config_Point->TmpDisk = Default_Record.Config.TmpDisk;
    Config_Point->Weight = Default_Record.Config.Weight;
    Config_Point->NodeBitMap = NULL;
    if (Default_Record.Config.Feature == (char *)NULL)
	Config_Point->Feature = (char *)NULL;
    else {
	Config_Point->Feature = (char *)malloc(strlen(Default_Record.Config.Feature)+1);
	if (Config_Point->Feature == (char *)NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Create_Config_Record: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Create_Config_Record: unable to allocate memory\n")
#endif
	    free(Config_Point);
	    *Error_Code = ENOMEM;
	    return (struct Config_Record *)NULL;
	} /* if */
	strcpy(Config_Point->Feature, Default_Record.Config.Feature);
    } /* else */

    if (list_append(Config_List, Config_Point) == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Create_Config_Record: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Create_Config_Record: unable to allocate memory\n")
#endif
	if (Config_Point->Feature != (char *)NULL) free(Config_Point->Feature);
	free(Config_Point);
	*Error_Code = ENOMEM;
	return (struct Config_Record *)NULL;
    } /* if */

    return Config_Point;
} /* Create_Config_Record */

/* 
 * Create_Node_Record - Create a node record
 * Input: Error_Code - Location to store error value in
 * Output: Error_Code - Set to zero if no error, errno otherwise
 *         Returns a pointer to the record or NULL if error
 * NOTE The record's values are initialized to those of Default_Record
 */
struct Node_Record *Create_Node_Record(int *Error_Code) {
    struct Node_Record *Node_Record_Point;
    int Old_Buffer_Size, New_Buffer_Size;

    *Error_Code = 0;

    /* Round up the buffer size to reduce overhead of realloc */
    Old_Buffer_Size = (Node_Record_Count) * sizeof(struct Node_Record);
    Old_Buffer_Size = ((int)((Old_Buffer_Size / NODE_BUF_SIZE) + 1)) * NODE_BUF_SIZE;
    New_Buffer_Size = (Node_Record_Count+1) * sizeof(struct Node_Record);
    New_Buffer_Size = ((int)((New_Buffer_Size / NODE_BUF_SIZE) + 1)) * NODE_BUF_SIZE;
    if (Node_Record_Count == 0)
	Node_Record_Table_Ptr = (struct Node_Record *)malloc(New_Buffer_Size);
    else if (Old_Buffer_Size != New_Buffer_Size)
	Node_Record_Table_Ptr = (struct Node_Record *)realloc(Node_Record_Table_Ptr, New_Buffer_Size);

    if (Node_Record_Table_Ptr == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Create_Node_Record: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Create_Node_Record: unable to allocate memory\n");
#endif
	*Error_Code = ENOMEM;
	return;
    } /* if */

    Node_Record_Point = Node_Record_Table_Ptr + (Node_Record_Count++);
    strcpy(Node_Record_Point->Name,    Default_Record.Node.Name);
    Node_Record_Point->NodeState     = Default_Record.Node.NodeState;
    Node_Record_Point->LastResponse  = (time_t) NULL;
    return Node_Record_Point;
} /* Create_Node_Record */


/* 
 * Create_Part_Record - Create a partition record
 * Input: Error_Code - Location to store error value in
 * Output: Error_Code - Set to zero if no error, errno otherwise
 *         Returns a pointer to the record or NULL if error
 * NOTE: The record's values are initialized to those of Default_Part
 */
struct Part_Record *Create_Part_Record(int *Error_Code) {
    struct Part_Record *Part_Record_Point;

    *Error_Code = 0;

    Part_Record_Point = (struct Part_Record *)malloc(sizeof(struct Part_Record));
    if (Part_Record_Point == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Create_Part_Record: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Create_Part_Record: unable to allocate memory\n");
#endif
	*Error_Code = ENOMEM;
	return (struct Part_Record *)NULL;
    } /* if */

    strcpy(Part_Record_Point->Name, "DEFAULT");
    Part_Record_Point->MaxTime     = Default_Part.MaxTime;
    Part_Record_Point->MaxNodes    = Default_Part.MaxNodes;
    Part_Record_Point->Key         = Default_Part.Key;
    Part_Record_Point->StateUp     = Default_Part.StateUp;
    Part_Record_Point->TotalNodes  = 0;
    Part_Record_Point->TotalCPUs   = 0;
    Part_Record_Point->NodeBitMap  = NULL;

    if (Default_Part.AllowGroups == (char *)NULL)
	Part_Record_Point->AllowGroups = (char *)NULL;
    else {
	Part_Record_Point->AllowGroups = (char *)malloc(strlen(Default_Part.AllowGroups)+1);
	if (Part_Record_Point->AllowGroups == NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Create_Part_Record: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Create_Part_Record: unable to allocate memory\n");
#endif
	    free(Part_Record_Point);
	    *Error_Code = ENOMEM;
	    return (struct Part_Record *)NULL;
	} /* if */
	strcpy(Part_Record_Point->AllowGroups, Default_Part.AllowGroups);
    } /* else */

    if (Default_Part.Nodes == (char *)NULL)
	Part_Record_Point->Nodes = (char *)NULL;
    else {
	Part_Record_Point->Nodes = (char *)malloc(strlen(Default_Part.Nodes)+1);
	if (Part_Record_Point->Nodes == NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Create_Part_Record: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Create_Part_Record: unable to allocate memory\n");
#endif
	    if (Part_Record_Point->AllowGroups != NULL) free(Part_Record_Point->AllowGroups);
	    free(Part_Record_Point);
	    *Error_Code = ENOMEM;
	    return NULL;
	} /* if */
	strcpy(Part_Record_Point->Nodes, Default_Part.Nodes);
    } /* else */

    if (list_append(Part_List, Part_Record_Point) == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Create_Part_Record: unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Create_Part_Record: unable to allocate memory\n")
#endif
	if (Part_Record_Point->Nodes       != NULL) free(Part_Record_Point->Nodes);
	if (Part_Record_Point->AllowGroups != NULL) free(Part_Record_Point->AllowGroups);
	free(Part_Record_Point);
	*Error_Code = ENOMEM;
	return (struct Part_Record *)NULL;
    } /* if */

    return Part_Record_Point;
} /* Create_Part_Record */


/* 
 * Delete_Node_Record - Delete record for node with specified name
 *   To avoid invalidating the bitmaps and hash table, we just clear the name 
 *   set its state to STATE_DOWN
 * Input: name - name of the desired node 
 * Output: return 0 on success, errno otherwise
 */
int Delete_Node_Record(char *name) {
    struct Node_Record *Node_Record_Point;	/* Pointer to Node_Record */

    Node_Record_Point = Find_Node_Record(name);
    if (Node_Record_Point == NULL) return ENOENT;

    strcpy(Node_Record_Point->Name, "");
    Node_Record_Point->NodeState = STATE_DOWN;
    return 0;
} /* Delete_Node_Record */


/* 
 * Delete_Part_Record - Delete record for partition with specified name
 * Input: name - name of the desired node 
 * Output: return 0 on success, errno otherwise
 */
int Delete_Part_Record(char *name) {
    ListIterator Part_Record_Iterator;		/* For iterating through Part_Record_List */
    struct Part_Record *Part_Record_Point;	/* Pointer to Part_Record */

    Part_Record_Iterator = list_iterator_create(Part_List);
    if (Part_Record_Iterator == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Delete_Part_Record: list_iterator_create unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Delete_Part_Record: list_iterator_create unable to allocate memory\n");
#endif
	return ENOMEM;
    } /* if */

    while (Part_Record_Point = (struct Part_Record *)list_next(Part_Record_Iterator)) {
	if (strcmp(Part_Record_Point->Name, name) == 0) continue;
	if (Part_Record_Point->AllowGroups) free(Part_Record_Point->AllowGroups);
	if (Part_Record_Point->Nodes)       free(Part_Record_Point->Nodes);
	free(Part_Record_Point);
	list_iterator_destroy(Part_Record_Iterator);
	return 0;
    } /* while */

    list_iterator_destroy(Part_Record_Iterator);
    return ENOENT;
} /* Delete_Part_Record */


/* Print the Hash_Table contents, used for debugging or analysis of hash technique */
void Dump_Hash() {
    int i;

    if (Hash_Table ==  NULL) return;
    for (i=0; i<Node_Record_Count; i++) {
	if (Hash_Table[i] == (int)NULL) continue;
	printf("Hash:%d:%s\n", i, (Node_Record_Table_Ptr+Hash_Table[i])->Name);
    } /* for */
} /* Dump_Hash */


/* 
 * Find_Node_Record - Find a record for node with specified name,
 * Input: name - name of the desired node 
 * Output: return pointer to node record or NULL if not found
 */
struct Node_Record *Find_Node_Record(char *name) {
    struct Node_Record *Node_Record_Point;	/* Pointer to Node_Record */
    int i;

    /* Try to find in hash table first */
    if (Hash_Table != NULL) {
	i = Hash_Index(name);
        if (strcmp((Node_Record_Table_Ptr+Hash_Table[i])->Name, name) == 0) 
		return (Node_Record_Table_Ptr+Hash_Table[i]);
#if DEBUG_SYSTEM
	fprintf(stderr, "Find_Node_Record: Hash table lookup failure for %s\n", name);
#else
	syslog(LOG_DEBUG, "Find_Node_Record: Hash table lookup failure for %s\n", name)
#endif
    } /* if */

#if DEBUG_SYSTEM
    if (Hash_Table != NULL) {
	printf("Sequential search for %s\n", name);
	Dump_Hash();
    } /* if */
#endif

    /* Revert to sequential search */
    for (i=0; i<Node_Record_Count; i++) {
	if (strcmp(name, (Node_Record_Table_Ptr+i)->Name) != 0) continue;
	return (Node_Record_Table_Ptr+i);
    } /* for */

    return (struct Node_Record *)NULL;
} /* Find_Node_Record */


/* 
 * Find_Part_Record - Find a record for node with specified name,
 * Input: name - name of the desired node 
 * Output: return pointer to node record or NULL if not found
 */
struct Part_Record *Find_Part_Record(char *name) {
    ListIterator Part_Record_Iterator;		/* For iterating through Part_Record_List */
    struct Part_Record *Part_Record_Point;	/* Pointer to Part_Record */

    Part_Record_Iterator = list_iterator_create(Part_List);
    if (Part_Record_Iterator == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Find_Part_Record: list_iterator_create unable to allocate memory\n");
#else
	syslog(LOG_ALERT, "Find_Part_Record: list_iterator_create unable to allocate memory\n");
#endif
	return NULL;
    }

    while (Part_Record_Point = (struct Part_Record *)list_next(Part_Record_Iterator)) {
	if (strcmp(Part_Record_Point->Name, name) == 0) break;
    } /* while */

    list_iterator_destroy(Part_Record_Iterator);
    return Part_Record_Point;	/* Value is NULL at end of list */
} /* Find_Part_Record */


/* 
 * Hash_Index - Return a hash table index for the given node name 
 * This code is optimized for names containing a base-ten suffix (e.g. "lx04")
 * Input: The node's name
 * Output: Return code is the hash table index
 */
int Hash_Index(char *name) {
    int i, inx, tmp;

    if (Node_Record_Count == 0) return 0;		/* Degenerate case */
    inx = 0;

#if ( HASH_BASE == 10 )
    for (i=0; ;i++) { 
	tmp = (int) name[i];
	if (tmp == 0) break;			/* end if string */
	if ((tmp >= (int)'0') && (tmp <= (int)'9')) 
	    inx = (inx * HASH_BASE) + (tmp - (int)'0');
    } /* for */
#elif ( HASH_BASE == 8 )
    for (i=0; ;i++) { 
	tmp = (int) name[i];
	if (tmp == 0) break;			/* end if string */
	if ((tmp >= (int)'0') && (tmp <= (int)'7')) 
	    inx = (inx * HASH_BASE) + (tmp - (int)'0');
    } /* for */

#else
    for (i=0; i<5;i++) { 
	tmp = (int) name[i];
	if (tmp == 0) break;					/* end if string */
	if ((tmp >= (int)'0') && (tmp <= (int)'9')) {		/* value 0-9 */
	    tmp -= (int)'0';
	} else if ((tmp >= (int)'a') && (tmp <= (int)'z')) {	/* value 10-35 */
	    tmp -= (int)'a';
	    tmp += 10;
	} else if ((tmp >= (int)'A') && (tmp <= (int)'Z')) {	/* value 10-35 */
	    tmp -= (int)'A';
	    tmp += 10;
	} else {
	    tmp = 36;
	}
	inx = (inx * 37) + tmp;
    } /* for */
 #endif

    inx = inx % Node_Record_Count;
    return inx;
} /* Hash_Index */


/* 
 * Init_SLURM_Conf - Initialize the SLURM configuration values. 
 * This should be called before first calling Read_SLURM_Conf.
 * Output: return value - 0 if no error, otherwise an error code
 */
int Init_SLURM_Conf() {

    BackupController = NULL;
    ControlMachine = NULL;

    strcpy(Default_Record.Node.Name, "DEFAULT");
    Default_Record.Node.NodeState    = STATE_UNKNOWN;
    Default_Record.Node.LastResponse = (time_t)0;
    Default_Record.Config.CPUs       = 1;
    Default_Record.Config.RealMemory = 1;
    Default_Record.Config.TmpDisk    = 1;
    Default_Record.Config.Weight     = 1;
    Default_Record.Config.Feature    = (char *)NULL;

    Node_Record_Table_Ptr = (struct Node_Record *)NULL;
    Node_Record_Count = 0;
    Hash_Table = (int *)NULL;
    Last_Node_Update = time(NULL);

    Config_List = list_create(NULL);
    if (Config_List == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Init_SLURM_Conf: list_create can not allocate memory\n");
#else
	syslog(LOG_ALARM, "Init_SLURM_Conf: list_create can not allocate memory\n");
#endif
	return ENOMEM;
    } /* if */

    Default_Part.AllowGroups = (char *)NULL;
    Default_Part.MaxTime     = -1;
    Default_Part.MaxNodes    = -1;
    Default_Part.Nodes       = (char *)NULL;
    Default_Part.Key         = 1;
    Default_Part.StateUp     = 1;

    Part_List = list_create(NULL);
    if (Part_List == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Init_SLURM_Conf: list_create can not allocate memory\n");
#else
	syslog(LOG_ALARM, "Init_SLURM_Conf: list_create can not allocate memory\n");
#endif
	return ENOMEM;
    } /* if */
    strcpy(Default_Part_Name, "");
    Default_Part_Loc = (struct Part_Record *)NULL;

    Up_NodeBitMap = Idle_NodeBitMap = NULL;
    return 0;
} /* Init_SLURM_Conf */


/*
 * Load_Integer - Parse a string for a keyword, value pair  
 * Input: *destination - Location into which result is stored
 *        keyword - String to search for
 *        In_Line - String to search for keyword
 * Output: *destination - set to value, No change if value not found, 
 *             Set to 1 if keyword found without value, 
 *             Set to -1 if keyword followed by "UNLIMITED"
 *         In_Line - The keyword and value (if present) are overwritten by spaces
 *         return value - 0 if no error, otherwise an error code
 */
int Load_Integer(int *destination, char *keyword, char *In_Line) {
    char Scratch[BUF_SIZE];	/* Scratch area for parsing the input line */
    char *str_ptr1, *str_ptr2, *str_ptr3;
    int i, str_len1, str_len2;

    str_ptr1 = (char *)strstr(In_Line, keyword);
    if (str_ptr1 != NULL) {
	str_len1 = strlen(keyword);
	strcpy(Scratch, str_ptr1+str_len1);
	if (isspace((int)Scratch[0])) { /* Keyword with no value set */
	    *destination = 1;
	    str_len2 = 0;
	} else {
	    str_ptr2 = (char *)strtok_r(Scratch, SEPCHARS, &str_ptr3);
	    str_len2 = strlen(str_ptr2);
	    if (strcmp(str_ptr2, "UNLIMITED") == 0)
		*destination = -1;
	    else if ((str_ptr2[0] >= '0') && (str_ptr2[0] <= '9')) 
		*destination = (int) strtol(Scratch, (char **)NULL, 10);
	    else {
#if DEBUG_SYSTEM
		fprintf(stderr, "Load_Integer: bad value for keyword %s\n", keyword);
#else
		syslog(LOG_ERR, "Load_Integer: bad value for keyword %s\n", keyword);
#endif
		return EINVAL;	
	    } /* else */
	} /* else */

	for (i=0; i<(str_len1+str_len2); i++) {
	    str_ptr1[i] = ' ';
	} /* for */
    } /* if */
    return 0;
} /* Load_Integer */


/*
 * Load_String - Parse a string for a keyword, value pair  
 * Input: *destination - Location into which result is stored
 *        keyword - String to search for
 *        In_Line - String to search for keyword
 * Output: *destination - set to value, No change if value not found, 
 *           this memory location must later be freed
 *	     if *destination had previous value, that memory location is automatically freed
 *         In_Line - The keyword and value (if present) are overwritten by spaces
 *         return value - 0 if no error, otherwise an error code
 */
int Load_String(char **destination, char *keyword, char *In_Line) {
    char Scratch[BUF_SIZE];	/* Scratch area for parsing the input line */
    char *str_ptr1, *str_ptr2, *str_ptr3;
    int i, str_len1, str_len2;

    str_ptr1 = (char *)strstr(In_Line, keyword);
    if (str_ptr1 != NULL) {
	str_len1 = strlen(keyword);
	strcpy(Scratch, str_ptr1+str_len1);
	if (isspace((int)Scratch[0])) { /* No value set */
#if DEBUG_SYSTEM
	    fprintf(stderr, "Load_String: keyword %s lacks value\n", keyword);
#else
	    syslog(LOG_ERR, "Load_String: keyword %s lacks value\n", keyword);
#endif
	    return EINVAL;
	} /* if */
	str_ptr2 = (char *)strtok_r(Scratch, SEPCHARS, &str_ptr3);
	str_len2 = strlen(str_ptr2);
	if (destination[0] != NULL) free(destination[0]);
	destination[0] = (char *)malloc(str_len2+1);
	if (destination[0] == NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Load_String: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Load_String: unable to allocate memory\n");
#endif
	    return ENOMEM;
	} /* if */
	strcpy(destination[0], str_ptr2);
	for (i=0; i<(str_len1+str_len2); i++) {
	    str_ptr1[i] = ' ';
	} /* for */
    } /* if */
    return 0;
} /* Load_String */


/* 
 * Parse_Node_Name - Parse the node name for regular expressions and return a sprintf format 
 * generate multiple node names as needed.
 * Input: NodeName - Node name to parse
 * Output: Format - sprintf format for generating names
 *         Start_Inx - First index to used
 *         End_Inx - Last index value to use
 *         Count_Inx - Number of index values to use (will be zero if none)
 *         return 0 if no error, error code otherwise
 * NOTE: The calling program must execute free(Format) when the storage location is no longer needed
 */
int Parse_Node_Name(char *NodeName, char **Format, int *Start_Inx, int *End_Inx, int *Count_Inx) {
    int Base, Format_Pos, Precision, i;
    char Type[1];

    i = strlen(NodeName);
    Format[0] = (char *)malloc(i+1);
    if (Format[0] == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Parse_Node_Name: unable to allocate memory\n");
#else
	syslog(LOG_ERR, "Parse_Node_Name: unable to allocate memory\n")
#endif
	return ENOMEM;
    } /* if */

    *Start_Inx = 0;
    *End_Inx   = 0;
    *Count_Inx = 0;
    Format_Pos = 0;
    Base = 0;
    Format[0][Format_Pos] = (char)NULL;
    i = 0;
    while (1) {
	if (NodeName[i] == (char)NULL) break;
	if (NodeName[i] == '\\') {
	    if (NodeName[++i] == (char)NULL) break;
	    Format[0][Format_Pos++] = NodeName[i++];
	} else if (NodeName[i] == '[') {
	    if (NodeName[++i] == (char)NULL) break;
	    if (Base != 0) {
#if DEBUG_SYSTEM
		fprintf(stderr, "Parse_Node_Name: Invalid '[' in node name %s\n", NodeName);
#else
		syslog(LOG_ALERT, "Parse_Node_Name: Invalid '[' in node name %s\n", NodeName)
#endif
		free(Format[0]);
		return EINVAL;
	    } /* if */
	    if (NodeName[i] == 'o') {
		Type[0] = NodeName[i++];
		Base = 8;
	    } else {
		Type[0] = 'd';
		Base = 10;
	    } /* else */
	    Precision = 0;
	    while (1) {
		if ((NodeName[i] >= '0') && (NodeName[i] <= '9')) {
		    *Start_Inx = ((*Start_Inx) * Base) + (int)(NodeName[i++] - '0');
		    Precision++;
		    continue;
		} /* if */
		if (NodeName[i] == '-') {
		    i++;
		    break;
		} /* if */
#if DEBUG_SYSTEM
		fprintf(stderr, "Parse_Node_Name: Invalid '%c' in node name %s\n", 
			NodeName[i], NodeName);
#else
		syslog(LOG_ALERT, "Parse_Node_Name: Invalid '%c' in node name %s\n", 
			NodeName[i], NodeName)
#endif
		free(Format[0]);
		return EINVAL;
	    } /* while */
	    while (1) {
		if ((NodeName[i] >= '0') && (NodeName[i] <= '9')) {
		    *End_Inx = ((*End_Inx) * Base) + (int)(NodeName[i++] - '0');
		    continue;
		} /* if */
		if (NodeName[i] == ']') {
		    i++;
		    break;
		} /* if */
#if DEBUG_SYSTEM
		fprintf(stderr, "Parse_Node_Name: Invalid '%c' in node name %s\n", 
			NodeName[i], NodeName);
#else
		syslog(LOG_ALERT, "Parse_Node_Name: Invalid '%c' in node name %s\n", 
			NodeName[i], NodeName)
#endif
		free(Format[0]);
		return EINVAL;
	    } /* while */
	    *Count_Inx = (*End_Inx - *Start_Inx) + 1;
	    Format[0][Format_Pos++] = '%';
	    Format[0][Format_Pos++] = '.';
	    if (Precision > 9) Format[0][Format_Pos++] = '0' + (Precision/10);
	    Format[0][Format_Pos++] = '0' + (Precision%10);
	    Format[0][Format_Pos++] = Type[0];
	} else {
	    Format[0][Format_Pos++] = NodeName[i++];
	} /* else */
    } /* while */
    Format[0][Format_Pos] = (char)NULL;
    return 0;
} /* Parse_Node_Name */


/* 
 * Parse_Node_Spec - Parse the node specification, build table and set values
 * Input:  In_Line line from the configuration file
 * Output: In_Line parsed keywords and values replaced by blanks
 *         return value 0 if no error, error code otherwise
 */
int Parse_Node_Spec (char *In_Line) {
    char *NodeName, *State, *Feature, *Format, This_Node_Name[BUF_SIZE];
    int Start_Inx, End_Inx, Count_Inx;
    int Error_Code, i;
    int State_Val, CPUs_Val, RealMemory_Val, TmpDisk_Val, Weight_Val;
    struct Node_Record *Node_Record_Point;
    struct Config_Record *Config_Point;
    char *str_ptr1, *str_ptr2;

    NodeName = State = Feature = (char *)NULL;
    CPUs_Val = RealMemory_Val = State_Val = NO_VAL;
    TmpDisk_Val = NO_VAL;
    Weight_Val = NO_VAL;
    if (Error_Code=Load_String(&NodeName, "NodeName=", In_Line))      return Error_Code;
    if (NodeName == NULL) return 0;	/* No Node info */

    Error_Code += Load_Integer(&CPUs_Val, "CPUs=", In_Line);
    Error_Code += Load_Integer(&RealMemory_Val, "RealMemory=", In_Line);
    Error_Code += Load_Integer(&TmpDisk_Val, "TmpDisk=", In_Line);
    Error_Code += Load_Integer(&Weight_Val, "Weight=", In_Line);
    if (Error_Code != 0) {
	free(NodeName);
	return Error_Code;
    } /* if */

    if (Error_Code=Load_String (&State, "State=", In_Line))           return Error_Code;
    if (State != NULL) {
	for (i=0; i<=STATE_END; i++) {
	    if (strcmp(Node_State_String[i], "END") == 0) break;
	    if (strcmp(Node_State_String[i], State) == 0) {
		State_Val = i;
		break;
	    } /* if */
	} /* for */
	if (State_Val == NO_VAL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Parse_Node_Spec: Invalid State %s for NodeName %s\n", State, NodeName);
#else
	    syslog(LOG_ERR, "Parse_Node_Spec: Invalid State %s for NodeName %s\n", State, NodeName);
#endif
	    free(NodeName);
	    free(State);
	    return EINVAL;
	} /* if */
	free(State);
    } /* if */

    Error_Code = Load_String (&Feature, "Feature=", In_Line);
    if (Error_Code != 0) {
	free(NodeName);
	return Error_Code;
    } /* if */

    Error_Code = Parse_Node_Name(NodeName, &Format, &Start_Inx, &End_Inx, &Count_Inx);
    if (Error_Code != 0) {
	free(NodeName);
	if (Feature) free(Feature);
	return Error_Code;
    } /* if */
    if (Count_Inx == 0) { 	/* Execute below loop once */
	Start_Inx = 0;
	End_Inx = 0;
    } /* if */

    for (i=Start_Inx; i<=End_Inx; i++) {
	if (Count_Inx == 0) {	/* Deal with comma separated node names here */
	    if (i == Start_Inx)
		str_ptr2 = strtok_r(Format, ",", &str_ptr1);
	    else
		str_ptr2 = strtok_r(NULL, ",", &str_ptr1);
	    if (str_ptr2 == NULL) break;
	    End_Inx++;
	    strcpy(This_Node_Name, str_ptr2);
	} else
	    sprintf(This_Node_Name, Format, i);
	if (strlen(This_Node_Name) >= MAX_NAME_LEN) {
#if DEBUG_SYSTEM
    	    fprintf(stderr, "Parse_Node_Spec: Node name %s too long\n", This_Node_Name);
#else
    	    syslog(LOG_ERR, "Parse_Node_Spec: Node name %s too long\n", This_Node_Name);
#endif
	    if (i != Start_Inx)  free(NodeName);
	    if (Format != NULL)  free(Format);
	    if (Feature != NULL) free(NodeName);
	    Error_Code = EINVAL;
	    break;
	} /* if */
	if (strcmp(This_Node_Name, "DEFAULT") == 0) {
	    if (CPUs_Val != NO_VAL)          Default_Record.Config.CPUs = CPUs_Val;
	    if (RealMemory_Val != NO_VAL)    Default_Record.Config.RealMemory = RealMemory_Val;
	    if (TmpDisk_Val != NO_VAL)       Default_Record.Config.TmpDisk = TmpDisk_Val;
	    if (Weight_Val != NO_VAL)        Default_Record.Config.Weight = Weight_Val;
	    if (State_Val != NO_VAL)         Default_Record.Node.NodeState = State_Val;
	    if (Feature != NULL) {
		if (Default_Record.Config.Feature) free(Default_Record.Config.Feature);
		Default_Record.Config.Feature = Feature;
	    } /* if */
	} else {
	    if (i == Start_Inx) {
		Config_Point = Create_Config_Record(&Error_Code);
		if (Error_Code != 0) {
		    free(NodeName);
		    if (Format)  free(Format);
		    if (Feature) free(Feature);
		    return Error_Code;
		} /* if */
		Config_Point->Nodes = NodeName;
		if (CPUs_Val != NO_VAL)       Config_Point->CPUs = CPUs_Val;
		if (RealMemory_Val != NO_VAL) Config_Point->RealMemory = RealMemory_Val;
		if (TmpDisk_Val != NO_VAL)    Config_Point->TmpDisk = TmpDisk_Val;
		if (Weight_Val != NO_VAL)     Config_Point->Weight = Weight_Val;
		if (Feature != NULL) {
		    if (Config_Point->Feature != NULL) free(Config_Point->Feature);
		    Config_Point->Feature = Feature;
		} /* if */
	    } /* if */

	    Node_Record_Point = Find_Node_Record(This_Node_Name);
	    if (Node_Record_Point == NULL) {
		Node_Record_Point = Create_Node_Record(&Error_Code);
		if (Error_Code != 0) break;
		strcpy(Node_Record_Point->Name, This_Node_Name);
		if (State_Val != NO_VAL)         Node_Record_Point->NodeState=State_Val;
		Node_Record_Point->Config_Ptr = Config_Point;
	    } else {
#if DEBUG_SYSTEM
		fprintf(stderr, "Parse_Node_Spec: Reconfiguration for node %s ignored.\n", 
		    This_Node_Name);
#else
		syslog(LOG_ERR, "Parse_Node_Spec: Reconfiguration for node %s ignored.\n", 
		    This_Node_Name);
#endif
	    } /* else */
	} /* else */
    } /* for (i */

    /* Free allocated storage */
    if (Format)  free(Format);
    return Error_Code;
} /* Parse_Node_Spec */


/*
 * Parse_Part_Spec - Parse the partition specification, build table and set values
 * Output: 0 if no error, error code otherwise
 */
int Parse_Part_Spec (char *In_Line) {
    int Line_Num;		/* Line number in input file */
    char *AllowGroups, *Nodes, *PartitionName, *State;
    int MaxTime_Val, MaxNodes_Val, Key_Val, Default_Val, StateUp_Val;
    int Error_Code, i;
    struct Part_Record *Part_Record_Point;

    AllowGroups = Nodes = PartitionName = State = (char *)NULL;
    MaxTime_Val = MaxNodes_Val = Key_Val = Default_Val = NO_VAL;

    if (Error_Code=Load_String(&PartitionName, "PartitionName=", In_Line))      return Error_Code;
    if (PartitionName == NULL) return 0;	/* No partition info */
	if (strlen(PartitionName) >= MAX_NAME_LEN) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Parse_Part_Spec: Partition name %s too long\n", PartitionName);
#else
	syslog(LOG_ERR, "Parse_Part_Spec: Partition name %s too long\n", PartitionName);
#endif
	free(PartitionName);
	return EINVAL;
    } /* if */

    Error_Code += Load_Integer(&MaxTime_Val,  "MaxTime=", In_Line);
    Error_Code += Load_Integer(&MaxNodes_Val, "MaxNodes=", In_Line);
    Error_Code += Load_Integer(&Default_Val,  "Default=NO", In_Line);
    if (Default_Val == 1) Default_Val=0;
    Error_Code += Load_Integer(&Default_Val,  "Default=YES", In_Line);
    Error_Code += Load_Integer(&StateUp_Val,  "State=DOWN", In_Line);
    if (StateUp_Val == 1) StateUp_Val=0;
    Error_Code += Load_Integer(&StateUp_Val,  "State=UP", In_Line);
    Error_Code += Load_Integer(&Key_Val,      "Key=NO", In_Line);
    if (Key_Val == 1) Key_Val=0;
    Error_Code += Load_Integer(&Key_Val,      "Key=YES", In_Line);
    if (Error_Code != 0) {
	free(PartitionName);
	return Error_Code;
    } /* if */

    Error_Code = Load_String (&Nodes, "Nodes=", In_Line);
    if (Error_Code != 0) {
	free(PartitionName);
	return Error_Code;
    } /* if */

    Error_Code = Load_String (&AllowGroups, "AllowGroups=", In_Line);
    if (Error_Code != 0) {
	free(PartitionName);
	if (Nodes != NULL) free(Nodes);
	return Error_Code;
    } /* if */

    if (strcmp(PartitionName, "DEFAULT") == 0) {
	if (MaxTime_Val  != NO_VAL)    Default_Part.MaxTime      = MaxTime_Val;
	if (MaxNodes_Val != NO_VAL)    Default_Part.MaxNodes     = MaxNodes_Val;
	if (Key_Val != NO_VAL)         Default_Part.Key          = Key_Val;
	if (StateUp_Val  != NO_VAL)    Default_Part.StateUp      = StateUp_Val;
	if (AllowGroups != (char *)NULL) {
	    if (Default_Part.AllowGroups != (char *)NULL) free(Default_Part.AllowGroups);
	    Default_Part.AllowGroups = AllowGroups;
	} /* if */
	if (Nodes != (char *)NULL) {
	    if (Default_Part.Nodes != (char *)NULL) free(Default_Part.Nodes);
	    Default_Part.Nodes = Nodes;
	} /* if */
	return 0;
    } /* if */

    Part_Record_Point = Find_Part_Record(PartitionName);
    if (Part_Record_Point == NULL) {
	Part_Record_Point = Create_Part_Record(&Error_Code);
	if (Error_Code != 0) {
	    free(PartitionName);
	    if (Nodes != NULL) free(Nodes);
	    if (AllowGroups != NULL) free(AllowGroups);
	    return Error_Code;
	} /* if */
	strcpy(Part_Record_Point->Name, PartitionName);
    } else {
#if DEBUG_SYSTEM
	fprintf(stderr, "Parse_Part_Spec: duplicate entry for partition %s\n", PartitionName);
#else
	syslog(LOG_NOTICE, "Parse_Node_Spec: duplicate entry for partition %s\n", PartitionName);
#endif
    } /* else */
    if (Default_Val  == 1) {
	if (strlen(Default_Part_Name) > 0) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Parse_Part_Spec: changing default partition from %s to %s\n", 
		    Default_Part_Name, PartitionName);
#else
	syslog(LOG_NOTICE, "Parse_Node_Spec: changing default partition from %s to %s\n", 
		    Default_Part_Name, PartitionName);
#endif
	} /* if */
	strcpy(Default_Part_Name, PartitionName);
	Default_Part_Loc = Part_Record_Point;
    } /* if */
    if (MaxTime_Val  != NO_VAL)    Part_Record_Point->MaxTime      = MaxTime_Val;
    if (MaxNodes_Val != NO_VAL)    Part_Record_Point->MaxNodes     = MaxNodes_Val;
    if (Key_Val  != NO_VAL)        Part_Record_Point->Key          = Key_Val;
    if (StateUp_Val  != NO_VAL)    Part_Record_Point->StateUp      = StateUp_Val;
    if (AllowGroups != (char *)NULL) {
	if (Part_Record_Point->AllowGroups != (char *)NULL) free(Part_Record_Point->AllowGroups);
	Part_Record_Point->AllowGroups = AllowGroups;
    } /* if */
    if (Nodes != (char *)NULL) {
	if (Part_Record_Point->Nodes != (char *)NULL) free(Part_Record_Point->Nodes);
	Part_Record_Point->Nodes = Nodes;
    } /* if */
    return 0;
} /* Parse_Part_Spec */


/*
 * Read_SLURM_Conf - Load the SLURM configuration from the specified file. 
 * Call Init_SLURM_Conf before ever calling Read_SLURM_Conf.  
 * Read_SLURM_Conf can be called more than once if so desired.
 * Input: File_Name - Name of the file containing overall SLURM configuration information
 * Output: return - 0 if no error, otherwise an error code
 */
int Read_SLURM_Conf (char *File_Name) {
    FILE *SLURM_Spec_File;	/* Pointer to input data file */
    int Line_Num;		/* Line number in input file */
    char In_Line[BUF_SIZE];	/* Input line */
    char Scratch[BUF_SIZE];	/* Scratch area for parsing the input line */
    char *str_ptr1, *str_ptr2, *str_ptr3;
    int i, j, Error_Code;

    /* Initialization */
    SLURM_Spec_File = fopen(File_Name, "r");
    if (SLURM_Spec_File == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Read_SLURM_Conf error %d opening file %s\n", errno, File_Name);
#else
	syslog(LOG_ALERT, "Read_SLURM_Conf error %d opening file %s\n", errno, File_Name);
#endif
	return errno;
    } /* if */

#if DEBUG_SYSTEM
    fprintf(stderr, "Read_SLURM_Conf: Loading configuration from %s\n", File_Name);
#else
    syslog(LOG_NOTICE, "Read_SLURM_Conf: Loading configuration from %s\n", File_Name);
#endif

    /* Process the data file */
    Line_Num = 0;
    while (fgets(In_Line, BUF_SIZE, SLURM_Spec_File) != NULL) {
	Line_Num++;
	if (strlen(In_Line) >= (BUF_SIZE-1)) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Read_SLURM_Conf line %d, of input file %s too long\n", 
		Line_Num, File_Name);
#else
	    syslog(LOG_ALERT, "Read_SLURM_Conf line %d, of input file %s too long\n", 
		Line_Num, File_Name);
#endif
	    return E2BIG;
	    break;
	} /* if */

	/* Everything after a non-escaped "#" is a comment */
	/* Replace comment flag "#" with an end of string (NULL) */
	for (i=0; i<BUF_SIZE; i++) {
	    if (In_Line[i] == (char)NULL) break;
	    if (In_Line[i] != '#') continue;
	    if ((i>0) && (In_Line[i-1]=='\\')) {	/* escaped "#" */
		for (j=i; j<BUF_SIZE; j++) {
		    In_Line[j-1] = In_Line[j];
		} /* for */
		continue;
	    } /* if */
	    In_Line[i] = (char)NULL;
	    break;
	} /* for */

	/* Parse what is left */
	/* Overall SLURM configuration parameters */
	if (Error_Code=Load_String(&ControlMachine, "ControlMachine=", In_Line))     return Error_Code;
	if (Error_Code=Load_String(&BackupController, "BackupController=", In_Line)) return Error_Code;

	/* Node configuration parameters */
	if (Error_Code=Parse_Node_Spec(In_Line)) return Error_Code;

	/* Partition configuration parameters */
	if (Error_Code=Parse_Part_Spec(In_Line)) return Error_Code;

	/* Report any leftover strings on input line */
	Report_Leftover(In_Line, Line_Num);
    } /* while */

    /* If values not set in configuration file, set defaults */
    if (BackupController == NULL) {
	BackupController = (char *)malloc(1);
	if (BackupController == NULL) {
#if DEBUG_SYSTEM
	    fprintf(stderr, "Read_SLURM_Conf: unable to allocate memory\n");
#else
	    syslog(LOG_ALERT, "Read_SLURM_Conf: unable to allocate memory\n")
#endif
	    return ENOMEM;
	} /* if */
	BackupController[0] = (char)NULL;
#if DEBUG_SYSTEM
    fprintf(stderr, "Read_SLURM_Conf: BackupController value not specified.\n");
#else
    syslog(LOG_WARN, "Read_SLURM_Conf: BackupController value not specified.\n")
#endif
    } /* if */
    if (ControlMachine == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Read_SLURM_Conf: ControlMachine value not specified.\n");
#else
	syslog(LOG_ALERT, "Read_SLURM_Conf: ControlMachine value not specified.\n")
#endif
	return EINVAL;
    } /* if */

    Rehash();
    if (Error_Code=Build_BitMaps()) return Error_Code;

#if DEBUG_SYSTEM
    fprintf(stderr, "Read_SLURM_Conf: Finished loading configuration\n");
#else
    syslog(LOG_NOTICE, "Read_SLURM_Conf: Finished loading configuration\n");
#endif

    return 0;
} /* Read_SLURM_Conf */


/* 
 * Rehash - Build a hash table of the Node_Record entries. This is a large hash table 
 * to permit the immediate finding of a record based only upon its name without regards 
 * to the number. There should be no need for a search. The algorithm is optimized for 
 * node names with a base-ten sequence number suffix. If you have a large cluster and 
 * use a different naming convention, this function and/or the Hash_Index function 
 * should be re-written.
 */
void Rehash() {
    struct Node_Record *Node_Record_Point;	/* Pointer to Node_Record */
    int i, inx;

    if (Hash_Table ==  (int *)NULL)
	Hash_Table = (int *)malloc(sizeof(int) * Node_Record_Count);
    else
	Hash_Table = (int *)realloc(Hash_Table, (sizeof(int) * Node_Record_Count));

    if (Hash_Table == NULL) {
#if DEBUG_SYSTEM
	fprintf(stderr, "Rehash: list_append can not allocate memory\n");
#else
	syslog(LOG_ERR, "Rehash: list_append can not allocate memory\n");
#endif
	return;
    } /* if */
    memset(Hash_Table, 0, (sizeof(int *) * Node_Record_Count));

    for (i=0; i<Node_Record_Count; i++) {
	if (strlen((Node_Record_Table_Ptr+i)->Name) == 0) continue;
	inx = Hash_Index((Node_Record_Table_Ptr+i)->Name);
	Hash_Table[inx] = i;
    } /* for */

} /* Rehash */


/* 
 * Report_Leftover - Report any un-parsed (non-whitespace) characters on the
 * configuration input line.
 * Input: In_Line - What is left of the configuration input line.
 *        Line_Num - Line number of the configuration file.
 * Output: NONE
 */
void Report_Leftover(char *In_Line, int Line_Num) {
    int Bad_Index, i;

    Bad_Index = -1;
    for (i=0; i<strlen(In_Line); i++) {
	if (In_Line[i] == '\n') In_Line[i]=' ';
	if (isspace((int)In_Line[i])) continue;
	Bad_Index=i;
	break;
    } /* if */

    if (Bad_Index == -1) return;
#if DEBUG_SYSTEM
    fprintf(stderr, "Report_Leftover: Ignored input on line %d of configuration: %s\n", 
	Line_Num, &In_Line[Bad_Index]);
#else
    syslog(LOG_ERR, "Report_Leftover: Ignored input on line %d of configuration: %s\n", 
	Line_Num, &In_Line[Bad_Index]);
#endif
    return;
} /* Report_Leftover */
